# import defusedxml.lxml
import urllib.parse

import xmltodict
from diff_match_patch import diff_match_patch
from lxml import etree  # nosec B410

# Safe XML parsing with custom options (solution of the author)
# https://github.com/tiran/defusedxml/issues/102
# Nb: From issue #33, defusedxml.lxml should be depracted by the author.
# PARSER = etree.Parser(remove_blank_text=True, resolve_entities=False)
# LOOKUP = etree.ElementDefaultClassLookup(defusedxml.lxml.RestrictedElement)
# PARSER.set_element_class_lookup(LOOKUP)

# https://stackoverflow.com/questions/3310614/remove-whitespaces-in-xml-string


def etree_fromstring(string, remove_blank_text=True):
    parser = etree.XMLParser(remove_blank_text=remove_blank_text)
    return etree.XML(string, parser=parser)


def etree_tostring(element, pretty_print=False):
    return etree.tostring(element, pretty_print=pretty_print)


def parse_response(response):
    if isinstance(response, (str, bytes)):
        # Nb: We should use defusedxml library but it doesn't support
        # Removing blank spaces
        response = etree_fromstring(response)
    data = response.xpath("/response/result/*")
    for d in data:
        detach(d)
    return data


def detach(e):
    parent = e.getparent()
    if parent is not None:
        parent.remove(e)
    return e


def delete_nat_membership(
    client,
):
    pass


def delete_policy_membership(element):
    entry = element.entry
    # TODO: Check type
    element.remove()  # Remove element from tree
    print(element.dumps(True))
    with entry.as_dict() as d:
        d.target.negate = "no"
        d["destination-hip"].member = "any"
    print(element.dumps(True))

    # client.update(e.xpath, e.dumps())
    # client.create(e.xpath, e.dumps())


def map_dicts(a, b):
    """
    Combine values from b with the value in a having the same key.
    """
    for uuid, u in a.items():
        r = b.get(uuid)
        if r is None:
            continue
        yield u, r


def extend_element(dest, elements):
    """
    Only add element that are not already in the destination
    element.extend(...) is causing duplicates entries because
    the merge is not controlled
    """
    children = {c.tag for c in dest.getchildren()}
    for e in elements:
        if e.tag in children:
            continue
        dest.append(e)
    return dest


def el2dict(e):
    return xmltodict.parse(etree_tostring(e))


# =============================================
# Utility function to compare the changes between configurations
# Nb: native difflib is too slow

# Using xmldiff.main.diff_trees gives false positives (can be filtered) and is very slow
# also, deepdiff.DeepDiff(el2dict(xml1), el2dict(xml2)).to_json() xpath is not able to use attributes to identify entries
# eval(re.findall(".*entry'\]\[\d+\]", diff["dictionary_item_added"][0])[0], {"root": y})


def _diff_patch(text1, text2):
    dmp = diff_match_patch()
    patches = dmp.patch_make(text1, text2)
    return urllib.parse.unquote(dmp.patch_toText(patches))


def diff_patch(xml1, xml2):
    if not isinstance(xml1, str):
        xml1 = etree_tostring(xml1).decode()
    if not isinstance(xml2, str):
        xml2 = etree_tostring(xml2).decode()
    return _diff_patch(xml1, xml2)



PATH_SEARCH_PREFIX = "./devices/entry/device-group"
CONFIG_XPATH = "/config/devices/entry/device-group"
PATH_NAME_PREFIX = "/config/devices/entry/device-group/entry"

# https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/get-started-with-the-pan-os-rest-api/pan-os-rest-api-error-codes
PANORAMA_ERRORS = {
    1: "The operation was canceled, typically by the caller.",
    2: "Unknown internal server error.",
    3: "Bad request. The caller specified an invalid parameter.",
    4: (
        "Gateway timeout. A firewall or Panorama module timed out "
        "before a backend operation completed."
    ),
    5: "Not found. The requested entity was not found.",
    6: "Conflict. The entity that the caller attempted to create already exists.",
    7: (
        "Forbidden. The caller does not have permission "
        "to execute the specified operation."
    ),
    8: "Resource exhausted. Some resource has been exhausted.",
    9: (
        "Failed precondition. The operation was rejected because the system"
        " is not in a state required for the execution of the operation."
    ),
    10: "Aborted because of conflict. A typical cause is a concurrency issue.",
    11: (
        "Out of range. The operation was attempted past a valid range."
        " And example is reaching an end-of-file."
    ),
    12: (
        "Not implemented. The operation is disabled,"
        " not implemented, or not supported."
    ),
    13: (
        "Internal server error. An unexpected and potentially"
        "serious internal error occurred."
    ),
    14: "Service unavailable. The service is temporarily unavailable.",
    15: ("Internal server error. Unrecoverable data loss or data corruption occurred."),
    16: (
        "Unauthorized. The request does not have valid authentication"
        "credentials to perform the operation."
    ),
}
SUCCESS_CODE = 19



import addict
import xmltodict
from lxml.etree import _Element  # nosec B410

from .utils import etree_fromstring, etree_tostring


# This code is safe using lxml
# It will only handle code already parsed
class Dict(addict.Dict):
    def __init__(self, xml=None):
        if isinstance(xml, _Element):
            xml = etree_tostring(xml)
        if isinstance(xml, str):
            if not xml:
                xml = "<entry/>"
            xml = xmltodict.parse(xml)
        return super().__init__(xml)

    def dumps(self):
        return xmltodict.unparse(self, full_document=False)

    def element(self):
        return etree_fromstring(self.dumps())



import logging
from itertools import product
from multiprocessing.pool import ThreadPool as Pool

import requests

from .constants import SUCCESS_CODE
from .utils import (
    detach,
    diff_patch,
    etree_fromstring,
    etree_tostring,
    extend_element,
    map_dicts,
)

log = logging.getLogger(__name__)


def get_tree(host, api_key, /, remove_blank_text=True, verify=False, timeout=None):
    res = requests.get(
        f"{host}/api?type=config&action=show&xpath=/config",
        headers={"X-PAN-KEY": api_key},
        verify=verify,
        timeout=timeout,
    )
    root_tree = etree_fromstring(res.content, remove_blank_text=remove_blank_text)
    try:
        tree = root_tree.xpath("/response/result/config")[0]
    except Exception:
        log.warning("Response doesn't contains the config tag. Response:")
        log.warning(etree_tostring(root_tree, pretty_print=True).decode()[:1000])
        raise Exception("Response doesn't contains the config tag. Check the logs")
    detach(tree)  # Remove the /response/result part
    return tree


def parse_msg_result(result):
    # sometimes, there is <line> elements inside of msg
    msg = "".join(result.xpath("./msg/text()"))
    if msg:
        return msg
    return "\n".join(result.xpath("./msg/line/text()"))


def _get_rule_use_cmd(device_group, position, rule_type, start_index, number):
    # positions = ("pre", "post")
    return f"""<show><policy-app>
        <mode>get-all</mode>
        <filter>(rule-state eq 'any')</filter>
        <vsysName>{device_group}</vsysName>
        <position>{position}</position>
        <type>{rule_type}</type>
        <anchor>{start_index}</anchor>
        <nrec>{number}</nrec>
        <pageContext>rule_usage</pageContext>
    </policy-app></show>"""


class XMLApi:
    def __init__(self, host, api_key, verify=False):
        self._host = host
        self._api_key = api_key
        self._url = f"{host}/api"
        self._verify = verify

    def _request(
        self,
        type,
        method="GET",
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        if verify is None:
            verify = self._verify
        headers = {"X-PAN-KEY": self._api_key}
        query_params = {"type": type}
        if params:
            query_params = {**query_params, **params}

        res = requests.request(
            method=method,
            url=self._url,
            params=query_params,
            headers=headers,
            verify=verify,
        )
        content = res.content.decode()
        tree = etree_fromstring(content, remove_blank_text=remove_blank_text)
        status = tree.attrib["status"]
        code = int(tree.get("code", SUCCESS_CODE))
        msg = parse_msg_result(tree)
        if status == "error" or code < SUCCESS_CODE:
            print(content[:500])
            raise Exception(msg)
        if msg:
            return msg
        return tree

    def _conf_request(
        self,
        xpath,
        action="get",
        method="GET",
        params={},
        remove_blank_text=True,
        verify=None,
    ):
        params = {"action": action, "xpath": xpath, **params}
        return self._request(
            "config",
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def _op_request(
        self,
        cmd,
        method="POST",
        params={},
        remove_blank_text=True,
        verify=None,
    ):
        params = {"cmd": cmd, **params}
        return self._request(
            "op",
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def _commit_request(
        self,
        cmd,
        method="POST",
        params={},
        remove_blank_text=True,
        verify=None,
    ):
        params = {"cmd": cmd, **params}
        return self._request(
            "commit",
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def configuration(
        self,
        xpath,
        action="get",
        method="GET",
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        return self._conf_request(
            "op",
            xpath,
            action=action,
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def operation(
        self,
        cmd,
        method="POST",
        params=None,
        remove_blank_text=True,
        verify=None,
    ):
        return self._op_request(
            "op",
            cmd,
            method=method,
            params=params,
            remove_blank_text=remove_blank_text,
            verify=verify,
        )

    def get_tree(self, extended=False, verify=None):
        if verify is None:
            verify = self._verify
        tree = get_tree(self._host, self._api_key, verify=verify)
        if extended:
            self._extend_tree_information(tree, verify=verify)
        return tree

    def _get_rule_use(self, device_group, position, rule_type, number=200, verify=None):
        results = []
        for i in range(100):
            cmd = _get_rule_use_cmd(
                device_group,
                position,
                rule_type,
                i * number,
                number,
            )
            res = self._op_request(cmd, verify=verify).xpath("result")[0]
            total_count = int(res.attrib["total-count"])
            results.extend(res.xpath("entry"))
            if len(results) >= total_count:
                break
        return results

    def get_rule_use(self, tree=None, max_threads=None, verify=None):
        if tree is None:
            tree = self.get_tree(verify=verify)
        device_groups = tree.xpath("devices/*/device-group/*/@name")
        positions = ("pre", "post")
        # rule_types = tuple({x.tag for x in tree.xpath(
        # "devices/*/device-group/*"
        # "/*[self::post-rulebase or self::pre-rulebase]/*")})
        rule_types = ("security", "pbf", "nat", "application-override")
        args_list = list(product(device_groups, positions, rule_types))

        def func(args):
            return self._get_rule_use(*args, verify=verify)

        threads = len(args_list)
        threads = min(max_threads or threads, threads)
        with Pool(len(args_list)) as pool:
            data = pool.map(func, args_list)
        return [entry for entry_list in data for entry in entry_list]

    def _get_rule_hit_count(self, device_group, rulebase, rule_type, verify=None):
        cmd = (
            "<show><rule-hit-count><device-group>"
            f"<entry name='{device_group}'><{rulebase}><entry name='{rule_type}'>"
            f"<rules><all/></rules></entry></{rulebase}></entry>"
            "</device-group></rule-hit-count></show>"
        )
        res = self._op_request(cmd, verify=verify)
        entries = res.xpath(".//rules/entry") or []
        # return entries
        return [(device_group, rulebase, rule_type, e) for e in entries]

    def get_rule_hit_count(self, tree=None, max_threads=None, verify=None):
        if tree is None:
            tree = self.get_tree(verify=verify)
        device_groups = tree.xpath("devices/*/device-group/*/@name")
        # rulebases = tuple({x.tag for x in tree.xpath(
        # "devices/*/device-group/*/*")})
        rulebases = ("pre-rulebase", "post-rulebase")
        # rule_types = tuple({x.tag for x in tree.xpath(
        # "devices/*/device-group/*"
        # "/*[self::post-rulebase or self::pre-rulebase]/*")})
        rule_types = ("security", "pbf", "nat", "application-override")
        args_list = list(product(device_groups, rulebases, rule_types))

        def func(args):
            return self._get_rule_hit_count(*args, verify=verify)

        threads = len(args_list)
        threads = min(max_threads or threads, threads)
        with Pool(len(args_list)) as pool:
            data = pool.map(func, args_list)
        return [entry for entry_list in data for entry in entry_list]

    def _extend_tree_information(
        self,
        tree,
        extended=None,
        max_threads=None,
        verify=None,
    ):
        if extended is None:
            extended = self.get_rule_use(tree, max_threads=max_threads, verify=verify)
        rules = tree.xpath(
            ".//device-group/entry/"
            "*[self::pre-rulebase or self::post-rulebase]/*/rules/entry[@uuid]",
        )
        ext_dict = {x.attrib.get("uuid"): x for x in extended}
        rules_dict = {x.attrib["uuid"]: x for x in rules}
        for ext, rule in map_dicts(ext_dict, rules_dict):
            extend_element(rule, ext)
            # rule.extend(ext) # This is causing duplicates entries
        return tree, extended

    def get(self, xpath, verify=None):
        """
        This will retrieve the xml definition based on the xpath
        The xpath doesn't need to be exact
        and can select multiple values at once.
        Still, it must at least speciy /config at is begining
        """
        return self._conf_request(xpath, action="show", method="GET", verify=verify)

    def delete(self, xpath, verify=None):
        """
        This will REMOVE the xml definition at the provided xpath.
        The xpath must be exact.
        """
        return self._conf_request(
            xpath,
            action="delete",
            method="DELETE",
            verify=verify,
        )

    def create(self, xpath, xml_definition, verify=None):
        """
        This will ADD the xml definition
        INSIDE the element at the provided xpath.
        The xpath must be exact.
        """
        # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/configuration-api/set-configuration
        params = {"element": xml_definition}
        return self._conf_request(
            xpath,
            action="set",
            method="POST",
            params=params,
            verify=verify,
        )

    def update(self, xpath, xml_definition, verify=None):
        """
        This will REPLACE the xml definition
        INSTEAD of the element at the provided xpath
        The xpath must be exact.
        Nb: We can pull the whole config, update it locally,
        and push the final result
        """
        # https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/pan-os-xml-api-request-types/configuration-api/set-configuration
        params = {"element": xml_definition}
        return self._conf_request(
            xpath,
            action="edit",
            method="POST",
            params=params,
            verify=verify,
        )

    def revert_changes(self, skip_validated=False, verify=None):
        skip = "<skip-validate>yes</skip-validate>" if skip_validated else ""
        cmd = f"<revert><config>{skip}</config></revert>"
        return self._op_request(cmd, verify=verify)

    def validate_changes(self, verify=None):
        cmd = "<validate><full></full></validate>"
        return self._op_request(cmd, verify=verify)

    def uncommited_changes(self, verify=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        cmd = "<show><config><list><changes></changes></list></config></show>"
        return self._op_request(cmd, verify=verify)

    def uncommited_changes_patch(self, verify=None):
        """
        Gives detailed information about pending changes
        (e.g. xpath, owner, action, ...)
        """
        candidate = self.candidate_config(verify=verify)
        running = self.running_config(verify=verify)
        return diff_patch(running, candidate)

    def uncommited_changes_summary(self, verify=None):
        """
        Only gives the concern device groups
        """
        cmd = "<show><config><list><change-summary/></list></config></show>"
        return self._op_request(cmd, verify=verify)

    def pending_changes(self, verify=None):
        """
        Result content is either 'yes' or 'no'
        """
        cmd = "<check><pending-changes></pending-changes></check>"
        return self._op_request(cmd, verify=verify)

    def candidate_config(self, verify=None):
        """
        Get the confirmation with pending changes
        """
        cmd = "<show><config><candidate></candidate></config></show>"
        return self._op_request(cmd, remove_blank_text=False, verify=verify)

    def running_config(self, verify=None):
        """
        Get the running configuration
        """
        cmd = "<show><config><running></running></config></show>"
        return self._op_request(cmd, remove_blank_text=False, verify=verify)

    def get_jobs(self, job_id=None, verify=None):
        cmd = "<show><jobs>{}</jobs></show>".format(
            f"<id>{job_id}</id>" if job_id else "<all></all>",
        )
        return self._op_request(cmd, verify=verify)

    def get_pending_jobs(self, verify=None):
        cmd = "<show><jobs><pending></pending></jobs></show>"
        return self._op_request(cmd, verify=verify)

    def commit_changes(self, force=False, verify=None):
        # force = "<force></force>"  # We don't want to support force option
        cmd = "<commit>{}</commit>".format("<force></force>" if force else "")
        return self._commit_request(cmd, verify=verify)



import logging

import click
from nagra_network_misc_utils.logger import set_default_logger

from .dump_dir import dump_dir

set_default_logger()
logging.getLogger().setLevel(logging.INFO)


@click.group()
def cli():
    pass


cli.add_command(dump_dir)

if __name__ == "__main__":
    cli()



import json
import logging
from itertools import chain

import requests

# Remove warning for unverified certificate
# https://stackoverflow.com/questions/27981545/suppress-insecurerequestwarning-unverified-https-request-is-being-made-in-pytho
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from .rest_resources import (
    PanoramaDevicesResourceType,
    PanoramaNetworkResourceType,
    PanoramaObjectsResourceType,
    PanoramaPanoramaResourceType,
    PanoramaPoliciesResourceType,
)

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

log = logging.getLogger(__name__)

# https://docs.paloaltonetworks.com/pan-os/9-1/pan-os-panorama-api/get-started-with-the-pan-os-rest-api/pan-os-rest-api-error-codes
PANORAMA_ERRORS = {
    1: ("The operation was canceled, typically by the caller."),
    2: ("Unknown internal server error."),
    3: ("Bad request. The caller specified an invalid parameter."),
    4: (
        "Gateway timeout. A firewall or Panorama module timed out "
        "before a backend operation completed."
    ),
    5: ("Not found. The requested entity was not found."),
    6: ("Conflict. The entity that the caller attempted to create already exists."),
    7: (
        "Forbidden. The caller does not have permission "
        "to execute the specified operation."
    ),
    8: ("Resource exhausted. Some resource has been exhausted."),
    9: (
        "Failed precondition. The operation was rejected because "
        "the system is not in a state required "
        "for the execution of the operation."
    ),
    10: ("Aborted because of conflict. A typical cause is a concurrency issue."),
    11: (
        "Out of range. The operation was attempted past a valid range. "
        "And example is reaching an end-of-file."
    ),
    12: (
        "Not implemented. The operation is disabled, not implemented, "
        "or not supported."
    ),
    13: (
        "Internal server error. An unexpected and potentially serious "
        "internal error occurred."
    ),
    14: ("Service unavailable. The service is temporarily unavailable."),
    15: ("Internal server error. Unrecoverable data loss or data corruption occurred."),
    16: (
        "Unauthorized. The request does not have "
        "valid authentication credentials to perform the operation."
    ),
}
SUCCESS_CODE = 19

OBJECT_RESOURCES = [
    "Addresses",
    "AddressGroups",
    "Regions",
    "Applications",
    "ApplicationGroups",
    "ApplicationFilters",
    "Services",
    "ServiceGroups",
    "Tags",
    "GlobalProtectHIPObjects",
    "GlobalProtectHIPProfiles",
    "ExternalDynamicLists",
    "CustomDataPatterns",
    "CustomSpywareSignatures",
    "CustomVulnerabilitySignatures",
    "CustomURLCategories",
    "AntivirusSecurityProfiles",
    "AntiSpywareSecurityProfiles",
    "VulnerabilityProtectionSecurityProfiles",
    "URLFilteringSecurityProfiles",
    "FileBlockingSecurityProfiles",
    "WildFireAnalysisSecurityProfiles",
    "DataFilteringSecurityProfiles",
    "DoSProtectionSecurityProfiles",
    "SecurityProfileGroups",
    "LogForwardingProfiles",
    "AuthenticationEnforcements",
    "DecryptionProfiles",
    "DecryptionForwardingProfiles",
    "Schedules",
    "SDWANPathQualityProfiles",
    "SDWANTrafficDistributionProfiles",
]

POLICY_RESOURCES = [
    "SecurityRules",
    "NATRules",
    "QoSRules",
    "PolicyBasedForwardingRules",
    "DecryptionRules",
    "TunnelInspectionRules",
    "ApplicationOverrideRules",
    "AuthenticationRules",
    "DoSRules",
    "SDWANRules",
]

NETWORK_RESOURCES = [
    "EthernetInterfaces",
    "AggregateEthernetInterfaces",
    "VLANInterfaces",
    "LoopbackInterfaces",
    "TunnelIntefaces",
    "SDWANInterfaces",
    "Zones",
    "VLANs",
    "VirtualWires",
    "VirtualRouters",
    "IPSecTunnels",
    "GRETunnels",
    "DHCPServers",
    "DHCPRelays",
    "DNSProxies",
    "GlobalProtectPortals",
    "GlobalProtectGateways",
    "GlobalProtectGatewayAgentTunnels",
    "GlobalProtectGatewaySatelliteTunnels",
    "GlobalProtectGatewayMDMServers",
    "GlobalProtectClientlessApps",
    "GlobalProtectClientlessAppGroups",
    "QoSInterfaces",
    "LLDP",
    "GlobalProtectIPSecCryptoNetworkProfiles",
    "IKEGatewayNetworkProfiles",
    "IKECryptoNetworkProfiles",
    "MonitorNetworkProfiles",
    "InterfaceManagementNetworkProfiles",
    "ZoneProtectionNetworkProfiles",
    "QoSNetworkProfiles",
    "LLDPNetworkProfiles",
    "SDWANInterfaceProfiles",
]

DEVICE_RESOURCES = [
    "VirtualSystems",
]

DEFAULT_PARAMS = {
    "output-format": "json",
}


class PanoramaAPI:
    def __init__(self, api_key=None, verbose=False, verify=False):
        self._verbose = verbose
        self._verify = verify
        self._api_key = api_key

    def _inner_request(
        self,
        method,
        url,
        params=None,
        headers=None,
        data=None,
        verify=None,
    ):
        if params is None:
            params = {}
        if headers is None:
            headers = {}
        if verify is None:
            verify = self._verify
        default_headers = {
            "X-PAN-KEY": self._api_key,
            # 'Accept': 'application/json, application/xml',
            # 'Content-Type': 'application/json'
        }
        headers = {**default_headers, **headers}
        params = {**DEFAULT_PARAMS, **params}
        res = requests.request(
            method,
            url,
            params=params,
            headers=headers,
            verify=verify,
        )
        # The API always returns a json, no matter what
        # if not res.ok:
        #     return None
        try:
            data = res.json()
            code = int(
                data.get("@code") or data.get("code") or SUCCESS_CODE,
            )  # Sometimes, the code is a string, some other times it is a int
            error_occured = (
                not res.ok
                or data.get("@status", "") != "success"
                # In case of success, the value 19 is used
                or code < SUCCESS_CODE
            )
            if not error_occured:
                return data, None
            message = (
                data.get("message")
                or PANORAMA_ERRORS.get(data["@code"])
                or "Something happened: " + json.dumps(data)
            )
            error = f"(CODE: {code}) {message}"
            if self._verbose:
                causes = list(
                    chain(
                        *(
                            details.get("causes", {})
                            for details in data.get("details", [])
                        ),
                    ),
                )
                details = "".join(c.get("description") for c in causes)
                error = f"{error} {details}"
            return data, error
        except Exception as e:
            return None, str(e)

    def _request(
        self,
        method,
        url,
        params=None,
        headers=None,
        data=None,
        verify=None,
        no_exception=False,
    ):
        data, error = (
            self._inner_request(
                method,
                url,
                params=params,
                headers=headers,
                data=data,
                verify=verify,
            )
            or {}
        )
        if error:
            if no_exception:
                log.error(f"Could not {method.lower()} {url}: {error}")
                return data, error
            raise Exception(error)
        data = data.get("result", {}).get("entry") or []
        return data, error

    def request(self, method, url, params=None, headers=None, data=None, verify=None):
        data, _ = (
            self._request(
                method,
                url,
                params=params,
                headers=headers,
                data=data,
                verify=verify,
            )
            or {}
        )
        return data

    def get(self, url, params=None, headers=None, data=None, verify=None):
        data, _ = (
            self._request(
                "GET",
                url,
                params=params,
                headers=headers,
                data=data,
                verify=verify,
            )
            or {}
        )
        return data
        # return data.get("result", {}).get("entry") or []

    def delete(self, url, params=None, headers=None, data=None, verify=None):
        data, _ = (
            self._request(
                "DELETE",
                url,
                params=params,
                headers=headers,
                data=data,
                verify=verify,
            )
            or {}
        )
        return data


class PanoramaClient:
    def __init__(
        self,
        domain,
        api_key=None,
        version="v10.1",
        verify=False,
        verbose=False,
    ):
        client = PanoramaAPI(api_key=api_key, verbose=verbose, verify=verify)
        self.client = client
        self.objects = PanoramaObjectsResourceType(client, domain, version=version)
        self.policies = PanoramaPoliciesResourceType(client, domain, version=version)
        self.network = PanoramaNetworkResourceType(client, domain, version=version)
        self.device = PanoramaDevicesResourceType(client, domain, version=version)
        self.panorama = PanoramaPanoramaResourceType(client, domain, version=version)



import json
import logging
from collections import defaultdict
from pathlib import Path

import click

# https://stackoverflow.com/questions/27981545/suppress-insecurerequestwarning-unverified-https-request-is-being-made-in-pytho
import urllib3

from .restapi import PanoramaClient
from .utils import el2dict
from .xmlapi import XMLApi

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
log = logging.getLogger("Dump")


def xpath(el, expr):
    res = el.xpath(expr)
    if not res:
        return None
    return res[0]


def missing_attributes(element):
    location = element.xpath("../../..")[0].tag or ""
    # Find the first ancestor below the device-group tag
    device_group = (element.xpath("ancestor::*[parent::device-group]/@name") or [""])[0]
    loc = device_group
    return {
        "@location": location,
        "@device-group": device_group,
        "@loc": loc,
    }


def transform_value(d, path, transform):
    if not path or not d:
        return d
    c = d
    for p in path[:-1]:
        c = c.get(p)
        if c is None:
            return d
    key = path[-1]
    value = c.get(key)
    res = transform(key, value)
    c[key] = res
    return d


def _wrap_list(value):
    if value is not None and not isinstance(value, list):
        value = [value]
    return value


def ensure_list(d, path):
    return transform_value(d, path, lambda _, v: _wrap_list(v))


# def convert_member(d, path):
#     """
#         Convert {member: ["...", "..."]} to [{"#text": "..."}]
#         at the position given by keys
#     """

#     def transform(_, value):
#         print('func', value)
#         return [{'#text': text} for text in _wrap_list(value['member'])]

#     return transform_value(d, path, transform)

# ============================================================
# Grouping utilities


def group_into(target, element, keys):
    if not keys:
        raise Exception("missing keys")
    for k in keys[:-1]:
        target = target.setdefault(k, {})
    target = target.setdefault(keys[-1], [])
    target.append(element)


def group_rules(rules):
    groups = {}
    for e in rules:
        dg = e["@device-group"]
        pre_post = e["@location"].split("-")[0]
        group_into(groups, e, (dg, pre_post))
    return groups


def _flatten_dict(d):
    if not isinstance(d, dict):
        yield (d,)  # Use a tuple because it will be unpacked
        return
    for k, v in d.items():
        tmp = list(_flatten_dict(v))
        for *keys, data in tmp:
            yield k, *keys, data


def flatten_grouped_rules(rules):
    "device-group -> pre/post -> rule"
    groups = group_rules(rules)
    return list(_flatten_dict(groups))


# ============================================================


def _convert_element_to_dict(element, func=None):
    d = el2dict(element)["entry"]
    d.update(missing_attributes(element))
    if func is not None:
        d = func(d)
    return d


def convert_elements_to_dict(elements, func=None):
    return [_convert_element_to_dict(e, func=func) for e in elements]


# ============================================================


def prepare_rules(tree, rule_definitions):
    for rule_def in rule_definitions:
        elements = tree.xpath(rule_def["xpath"])
        data = convert_elements_to_dict(elements)
        rule_def["data"] = data
    return rule_definitions


def _dump_rules(directory, rule_type, rules):
    if isinstance(directory, str):
        directory = Path(directory).resolve()
    directory.mkdir(parents=True, exist_ok=True)
    # device-group -> pre/post -> rule
    grouped = flatten_grouped_rules(rules)
    for args in grouped:
        try:
            device_group, pre_post, rules = args
            # TODO: FIX => rules can be a mere dictionnary sometimes
        except Exception:
            print(args)
            raise
        filename = f"{pre_post}{rule_type}RulesFW{device_group}.json"
        with open(directory / filename, "w") as f:
            json.dump(rules, f)


def dump_rules(output_dir, rules):
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    for rule_def in rules:
        _dump_rules(
            output_dir / rule_def["folder"],
            rule_def["rule_type"],
            rule_def["data"],
        )


# ============================================================

RULES = [
    {
        "rule_type": "security",
        "folder": "securityRules",
        "xpath": (
            "devices/entry/device-group/*"
            "/*[self::post-rulebase or self::pre-rulebase]/security/rules/entry"
        ),
    },
    {
        "rule_type": "nat",
        "folder": "nat",
        "xpath": (
            "devices/entry/device-group/*"
            "/*[self::post-rulebase or self::pre-rulebase]/nat/rules/entry"
        ),
    },
    {
        "rule_type": "pbf",
        "folder": "pb",
        "xpath": (
            "devices/entry/device-group/*"
            "/*[self::post-rulebase or self::pre-rulebase]/pbf/rules/entry"
        ),
    },
    {
        "rule_type": "appover",
        "folder": "appover",
        "xpath": (
            "devices/entry/device-group/*"
            "/*[self::post-rulebase or self::pre-rulebase]"
            "/application-override/rules/entry"
        ),
    },
]


def dump_dir_xml(output, host, api_key):
    client = XMLApi(host, api_key)
    log.info("Starting to retrieve the data (extended informations)")
    tree = client.get_tree(extended=True)

    log.info("Data retrieved, starting the export")
    output = Path(output).resolve()
    output.mkdir(parents=True, exist_ok=True)

    address_elements = tree.xpath("devices/entry/device-group/*/address/entry")
    address_group_elements = tree.xpath(
        "devices/entry/device-group/*/address-group/entry",
    )

    addresses = convert_elements_to_dict(
        address_elements,
        lambda d: ensure_list(d, ("tag", "member")),
    )
    addresses_groups = convert_elements_to_dict(
        address_group_elements,
        lambda d: ensure_list(d, ("static", "member")),
    )

    with open(output / "addresses.json", "w") as f:
        json.dump(addresses, f)
    with open(output / "addressGroups.json", "w") as f:
        json.dump(addresses_groups, f)

    prepare_rules(tree, RULES)
    dump_rules(output, RULES)
    log.info("Export done")


def get_rest_rules(client, device_group):
    return [
        {
            "rule_type": "security",
            "folder": "securityRules",
            "pre-data": client.policies.SecurityPreRules.get(
                device_group=device_group,
                inherited=False,
            ),
            "post-data": client.policies.SecurityPostRules.get(
                device_group=device_group,
                inherited=False,
            ),
        },
        {
            "rule_type": "nat",
            "folder": "nat",
            "pre-data": client.policies.NATPreRules.get(
                device_group=device_group,
                inherited=False,
            ),
            "post-data": client.policies.NATPostRules.get(
                device_group=device_group,
                inherited=False,
            ),
        },
        {
            "rule_type": "pbf",
            "folder": "pb",
            "pre-data": client.policies.PolicyBasedForwardingPreRules.get(
                device_group=device_group,
                inherited=False,
            ),
            "post-data": client.policies.PolicyBasedForwardingPostRules.get(
                device_group=device_group,
                inherited=False,
            ),
        },
        {
            "rule_type": "appover",
            "folder": "appover",
            "pre-data": client.policies.ApplicationOverridePreRules.get(
                device_group=device_group,
                inherited=False,
            ),
            "post-data": client.policies.ApplicationOverridePostRules.get(
                device_group=device_group,
                inherited=False,
            ),
        },
    ]


def grouped_rest_rules(rules):
    # device-group -> pre/post -> rule
    grouped = {}
    for r in rules:
        folder = r["folder"]
        rule_type = r["rule_type"]
        for pre_post in ("pre", "post"):
            for e in r[pre_post + "-data"]:
                dg = e["@device-group"]
                group_into(grouped, e, (folder, rule_type, dg, pre_post))
    grouped = _flatten_dict(grouped)
    data = defaultdict(list)
    for args in grouped:
        try:
            folder, rule_type, device_group, pre_post, rules = args
            # TODO: FIX => rules can be a mere dictionnary sometimes
        except Exception:
            print(args)
            raise
        filename = f"{pre_post}{rule_type}RulesFW{device_group}.json"
        data[folder].append((filename, rules))
    return dict(data)


def dump_rest_rules(output_dir, rules):
    output_dir = Path(output_dir)
    for folder, data in grouped_rest_rules(rules).items():
        directory = output_dir / folder
        directory.mkdir(parents=True, exist_ok=True)
        for filename, rules in data:
            with open(directory / filename, "w") as f:
                json.dump(rules, f)


def dump_dir_rest(output, host, api_key):
    client = PanoramaClient(host, api_key)
    log.info("Starting to retrieve the data (extended informations)")
    groups = [d["@name"] for d in client.panorama.DeviceGroups.get()]
    addresses = client.objects.Addresses.get(device_group=groups, inherited=False)
    addresses_groups = client.objects.AddressGroups.get(
        device_group=groups,
        inherited=False,
    )

    rules = get_rest_rules(client, groups)
    log.info("Data retrieved, starting the export")
    output = Path(output).resolve()
    output.mkdir(parents=True, exist_ok=True)

    with open(output / "addresses.json", "w") as f:
        json.dump(addresses, f)
    with open(output / "addressGroups.json", "w") as f:
        json.dump(addresses_groups, f)
    dump_rest_rules(output, rules)
    log.info("Export done")


@click.command(
    "dump-dir",
    help="Dump data from palo alto",
)
@click.option("-o", "--output", help="Output directory", default="output")
@click.option(
    "-h",
    "--host",
    envvar="PANO_HOST",
    help="Palo Alto host (default to PANO_HOST env var)",
    required=True,
)
@click.option(
    "-k",
    "--key",
    "api_key",
    envvar="PANO_API_KEY",
    help="Palo Alto API Key (default to PANO_API_KEY env var)",
    required=True,
)
@click.option("--api", type=click.Choice(["xml", "rest"]), default="rest")
def dump_dir(output, host, api_key, api):
    host = host.strip()  # remove trailing \r
    api_key = api_key.strip()  # remove trailing \r
    if api == "xml":
        dump_dir_xml(output, host, api_key)
    else:
        dump_dir_rest(output, host, api_key)



from contextlib import contextmanager
from typing import ClassVar

from .constants import PATH_NAME_PREFIX, PATH_SEARCH_PREFIX
from .utils import detach, etree_tostring
from .wrappers import Dict


def _select_xpath_unique_selector(element):
    # identifiers = ("name", "uuid")
    identifiers = ("uuid", "name")
    for i in identifiers:
        value = element.get(i)
        if value:
            return i, value
    return None, None


def mk_xpath_selector(key, value):
    return f"[@{key}='{value}']"


def xpath_unique_selector(element):
    key, value = _select_xpath_unique_selector(element)
    if key:
        return mk_xpath_selector(key, value)
    # text = element.text.strip()
    # if text:
    #     return "[text()='{}']".format(text)
    return ""


def get_element_xpath(element, unambiguous=False):
    """
    This function return the xpath of an element relatively
    to the root. e.g. /config/devices/...

    unambiguous: Ensure that every possible node
        of the path uses its identifier
    """
    parts = []
    elements = [element, *list(element.iterancestors())]
    selector_ok = False
    for el in elements:
        tag = el.tag
        if not selector_ok or unambiguous:
            selector = xpath_unique_selector(el)
            if selector:
                selector_ok = True
                tag += selector
        parts.append(tag)
    return "/" + "/".join(parts[::-1])


def dump_element(element, pretty=True):
    return etree_tostring(element, pretty_print=pretty).decode()


def find_by_attr(tree, key, value):
    if not isinstance(value, str) or "'" in value:
        return None
    xpath = f"{PATH_SEARCH_PREFIX}//*[@{key} = '{value}']"
    return tree.xpath(xpath)


def find_by_name(tree, object_name):
    return find_by_attr(tree, "name", object_name)


def find_by_uuid(tree, uuid):
    return find_by_attr(tree, "uuid", uuid)


# this will consider all tags:
#   member, translation-address, ... be cautious using it
def find_references(tree, object_name, tag="*"):
    if not isinstance(tag, str):
        # Combine a list of tag into one
        # //*[self::tag1 or self::tag2]
        # //(tag1|tag2)
        tag = "({})".format("|".join(tag))
    if not isinstance(object_name, str) or "'" in object_name:
        return []
    return tree.xpath(
        f"{PATH_SEARCH_PREFIX}//{tag}[text() = '{object_name}']",
    )


def find_ancestor(element, tag):
    return next((x for x in element.iterancestors() if x.tag == tag), None)


class ConfigElement:
    def __init__(self, el, config=None, xpath=None):
        if config is None:
            config = Config(list(el.iterancestors())[-1])
        if xpath is None:
            xpath = get_element_xpath(el)
        self._el = el
        self._config = config
        self._xpath = xpath

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.xpath.replace(PATH_NAME_PREFIX, "")

    def dumps(self, pretty=False):
        return etree_tostring(self._el, pretty_print=pretty).decode()

    @property
    def element(self):
        return self._el

    @property
    def attrib(self):
        return self._el.attrib

    @property
    def name(self):
        return self._el.attrib.get("name")

    @property
    def config(self):
        return self._config

    @property
    def entry(self):
        e = find_ancestor(self._el, "entry")
        return self._wrap_element(e)

    @property
    def parent(self):
        e = self._el.getparent()
        if e is None:
            return None
        return self._wrap_element(e)

    @contextmanager
    def as_dict(self):
        d = Dict(self._el)
        try:
            yield d
        finally:
            new_element = d.element()
            parent = self._el.getparent()
            if parent:
                parent.replace(self._el, new_element)
            self._el = new_element
            self._set_xpath()

    def xpath(self, search=None):
        if search:
            e = self._el.xpath(search)
            return self._wrap_element(e)
        return self._xpath

    def _set_xpath(self, unambiguous=True):
        self._xpath = self.get_xpath(unambiguous=unambiguous)

    def get_xpath(self, unambiguous=True):
        return get_element_xpath(self._el, unambiguous=unambiguous)

    def _wrap_element(self, element):
        if element is None:
            return None
        return ConfigElement(element, self.config)

    def get_references(self):
        references = find_references(self.config.tree, self.name)
        return [self._wrap_element(e) for e in references]

    def remove(self):
        detach(self._el)


class Config:
    # Singleton like: Always return the same Config object for the same tree
    __instances: ClassVar = {}

    def __new__(cls, config_tree, *args, **kwargs):
        instance = cls.__instances.get(config_tree)
        if not instance:
            instance = object.__new__(cls)
            cls.__instances[config_tree] = instance
        return instance

    def __init__(self, config_tree):
        if config_tree.tag != "config":
            raise Exception("Wront tree, root must be <config>")
        detach(config_tree)
        self._tree = config_tree

    # def __str__(self):
    #     return etree_tostring(self._tree).decode()

    def dumps(self, pretty=False):
        return etree_tostring(self._tree, pretty_print=pretty).decode()

    @property
    def tree(self):
        return self._tree

    def _mk_config_el(self, el):
        return ConfigElement(el, self)

    def find_by_name(self, name):
        res = find_by_name(self._tree, name)
        return [self._mk_config_el(x) for x in res]

    def find_by_uuid(self, name):
        res = find_by_uuid(self._tree, name)
        return [self._mk_config_el(x) for x in res]



from functools import wraps
from multiprocessing.pool import ThreadPool as Pool
from typing import ClassVar

# List of existing resources
# https://docs.paloaltonetworks.com/pan-os/10-2/pan-os-panorama-api/get-started-with-the-pan-os-rest-api/access-the-rest-api

DEFAULT_PARAMS = {
    "location": "device-group",
}


class BasePanoramaResource:
    def __init__(
        self,
        client,
        host,
        resource_type,
        resource,
        version="v10.1",
        default_params=None,
    ):
        self.DEFAULT_PARAMS = default_params or DEFAULT_PARAMS
        if not host.startswith("http"):
            host = "https://" + host
        self._client = client
        self._name = f"{resource_type}/{resource}"
        self._url = f"{host}/restapi/{version}/{resource_type}/{resource}"

    def __str__(self):
        return self._name

    def __repr__(self):
        return str(self)

    def get(self, params=None, headers=None, data=None, verify=None):
        return self._client.get(
            self._url,
            params=params,
            headers=headers,
            data=data,
            verify=verify,
        )

    def delete(self, params=None, headers=None, data=None, verify=None):
        return self._client.delete(
            self._url,
            params=params,
            headers=headers,
            data=data,
            verify=verify,
        )


class PanoramaResource(BasePanoramaResource):
    def get(
        self,
        device_group=None,
        name=None,
        inherited=True,
        params=None,
        headers=None,
        data=None,
        verify=None,
    ):
        if params is None:
            params = {}
        _params = {**self.DEFAULT_PARAMS, **params}
        if name is not None:
            _params["name"] = name

        def func(params):
            params = {**_params, **params}
            res = self._client.get(
                self._url,
                params=params,
                headers=headers,
                data=data,
                verify=verify,
            )
            if not isinstance(res, list):
                return []  # Error happened
            if not inherited:
                res = [r for r in res if r["@device-group"] == r["@loc"]]
            return res

        if isinstance(device_group, str):
            device_group = [device_group]
        if len(device_group) == 1:
            params = {"device-group": device_group}
            return func(params)
        if len(device_group) > 1:
            params_list = [{"device-group": dg} for dg in device_group]
            with Pool(len(device_group)) as pool:
                data = pool.map(func, params_list)
            return [record for res in data for record in res]
        return None

    def delete(
        self,
        device_group=None,
        name=None,
        params=None,
        headers=None,
        data=None,
        verify=None,
    ):
        if params is None:
            params = {}
        params = {**self.DEFAULT_PARAMS, **params}
        if device_group is not None:
            params["device-group"] = device_group
        if name is not None:
            params["name"] = name
        return self._client.delete(
            self._url,
            params=params,
            headers=headers,
            data=data,
            verify=verify,
        )


class GroupedResources:
    def __init__(self, *resources):
        self._resources = resources

    def _grouped(self, method, *args, **kwargs):
        params_list = [(getattr(r, method), args, kwargs) for r in self._resources]

        def func(args):
            f, args, kwargs = args
            return f(*args, **kwargs)

        with Pool(len(self._resources)) as pool:
            data = pool.map(func, params_list)
        return [record for res in data for record in res]

    @wraps(PanoramaResource.get)
    def get(self, *args, **kwargs):
        return self._grouped("get", *args, **kwargs)

    # def delete(self, params={}, headers={}, data=None, verify=None):
    #     return self._client.delete(
    #         self._url,
    #         params=params, headers=headers,
    #         data=data, verify=verify
    #     )


class PanoramaResourceType:
    def __init__(self, client, resource_type):
        self._client = client
        self._name = resource_type

    def __str__(self):
        return self._name

    def __repr__(self):
        return str(self)


class PanoramaPanoramaResourceType(PanoramaResourceType):
    resource_type = "Panorama"

    def __init__(self, client, domain, version="v10.1"):
        resource_type = self.resource_type
        super().__init__(client, resource_type)

        def resource(name):
            return BasePanoramaResource(
                client,
                domain,
                resource_type,
                name,
                version=version,
            )

        self.DeviceGroups = resource("DeviceGroups")


class PanoramaObjectsResourceType(PanoramaResourceType):
    resource_type = "Objects"

    def __init__(self, client, domain, version="v10.1"):
        resource_type = self.resource_type
        super().__init__(client, resource_type)

        def resource(name):
            return PanoramaResource(
                client,
                domain,
                resource_type,
                name,
                version=version,
            )

        self.Addresses = resource("Addresses")
        self.AddressGroups = resource("AddressGroups")
        self.Regions = resource("Regions")
        self.Applications = resource("Applications")
        self.ApplicationGroups = resource("ApplicationGroups")
        self.ApplicationFilters = resource("ApplicationFilters")
        self.Services = resource("Services")
        self.ServiceGroups = resource("ServiceGroups")
        self.Tags = resource("Tags")
        self.GlobalProtectHIPObjects = resource("GlobalProtectHIPObjects")
        self.GlobalProtectHIPProfiles = resource("GlobalProtectHIPProfiles")
        self.ExternalDynamicLists = resource("ExternalDynamicLists")
        self.CustomDataPatterns = resource("CustomDataPatterns")
        self.CustomSpywareSignatures = resource("CustomSpywareSignatures")
        self.CustomVulnerabilitySignatures = resource("CustomVulnerabilitySignatures")
        self.CustomURLCategories = resource("CustomURLCategories")
        self.AntivirusSecurityProfiles = resource("AntivirusSecurityProfiles")
        self.AntiSpywareSecurityProfiles = resource("AntiSpywareSecurityProfiles")
        self.VulnerabilityProtectionSecurityProfiles = resource(
            "VulnerabilityProtectionSecurityProfiles",
        )
        self.URLFilteringSecurityProfiles = resource("URLFilteringSecurityProfiles")
        self.FileBlockingSecurityProfiles = resource("FileBlockingSecurityProfiles")
        self.WildFireAnalysisSecurityProfiles = resource(
            "WildFireAnalysisSecurityProfiles",
        )
        self.DataFilteringSecurityProfiles = resource("DataFilteringSecurityProfiles")
        self.DoSProtectionSecurityProfiles = resource("DoSProtectionSecurityProfiles")
        self.SecurityProfileGroups = resource("SecurityProfileGroups")
        self.LogForwardingProfiles = resource("LogForwardingProfiles")
        self.AuthenticationEnforcements = resource("AuthenticationEnforcements")
        self.DecryptionProfiles = resource("DecryptionProfiles")
        self.DecryptionForwardingProfiles = resource("DecryptionForwardingProfiles")
        self.Schedules = resource("Schedules")
        self.SDWANPathQualityProfiles = resource("SDWANPathQualityProfiles")
        self.SDWANTrafficDistributionProfiles = resource(
            "SDWANTrafficDistributionProfiles",
        )


class PanoramaPoliciesResourceType(PanoramaResourceType):
    resource_type = "Policies"

    def __init__(self, client, domain, version="v10.1"):
        resource_type = self.resource_type
        super().__init__(client, resource_type)

        def resource(name):
            return PanoramaResource(
                client,
                domain,
                resource_type,
                name,
                version=version,
            )

        self.SecurityPreRules = resource("SecurityPreRules")
        self.SecurityPostRules = resource("SecurityPostRules")
        self.SecurityRules = GroupedResources(
            self.SecurityPreRules,
            self.SecurityPostRules,
        )
        self.NATPreRules = resource("NATPreRules")
        self.NATPostRules = resource("NATPostRules")
        self.NATRules = GroupedResources(self.NATPreRules, self.NATPostRules)
        self.QoSPreRules = resource("QoSPreRules")
        self.QoSPostRules = resource("QoSPostRules")
        self.QoSRules = GroupedResources(self.QoSPreRules, self.QoSPostRules)
        self.PolicyBasedForwardingPreRules = resource("PolicyBasedForwardingPreRules")
        self.PolicyBasedForwardingPostRules = resource("PolicyBasedForwardingPostRules")
        self.PolicyBasedForwardingRules = GroupedResources(
            self.PolicyBasedForwardingPreRules,
            self.PolicyBasedForwardingPostRules,
        )
        self.DecryptionPreRules = resource("DecryptionPreRules")
        self.DecryptionPostRules = resource("DecryptionPostRules")
        self.DecryptionRules = GroupedResources(
            self.DecryptionPreRules,
            self.DecryptionPostRules,
        )
        self.TunnelInspectionPreRules = resource("TunnelInspectionPreRules")
        self.TunnelInspectionPostRules = resource("TunnelInspectionPostRules")
        self.TunnelInspectionRules = GroupedResources(
            self.TunnelInspectionPreRules,
            self.TunnelInspectionPostRules,
        )
        self.ApplicationOverridePreRules = resource("ApplicationOverridePreRules")
        self.ApplicationOverridePostRules = resource("ApplicationOverridePostRules")
        self.ApplicationOverrideRules = GroupedResources(
            self.ApplicationOverridePreRules,
            self.ApplicationOverridePostRules,
        )
        self.AuthenticationPreRules = resource("AuthenticationPreRules")
        self.AuthenticationPostRules = resource("AuthenticationPostRules")
        self.AuthenticationRules = GroupedResources(
            self.AuthenticationPreRules,
            self.AuthenticationPostRules,
        )
        self.DoSPreRules = resource("DoSPreRules")
        self.DoSPostRules = resource("DoSPostRules")
        self.DoSRules = GroupedResources(self.DoSPreRules, self.DoSPostRules)
        self.SDWANPreRules = resource("SDWANPreRules")
        self.SDWANPostRules = resource("SDWANPostRules")
        self.SDWANRules = GroupedResources(self.SDWANPreRules, self.SDWANPostRules)


class PanoramaNetworkResourceType(PanoramaResourceType):
    DEFAULT_PARAMS: ClassVar = {
        "location": "template",
    }
    resource_type = "Network"

    def __init__(self, client, domain, version="v10.1"):
        resource_type = self.resource_type
        super().__init__(client, resource_type)

        def resource(name):
            return PanoramaResource(
                client,
                domain,
                resource_type,
                name,
                version=version,
                default_params=self.DEFAULT_PARAMS,
            )

        self.EthernetInterfaces = resource("EthernetInterfaces")
        self.AggregateEthernetInterfaces = resource("AggregateEthernetInterfaces")
        self.VLANInterfaces = resource("VLANInterfaces")
        self.LoopbackInterfaces = resource("LoopbackInterfaces")
        self.TunnelIntefaces = resource("TunnelIntefaces")
        self.SDWANInterfaces = resource("SDWANInterfaces")
        self.Zones = resource("Zones")
        self.VLANs = resource("VLANs")
        self.VirtualWires = resource("VirtualWires")
        self.VirtualRouters = resource("VirtualRouters")
        self.IPSecTunnels = resource("IPSecTunnels")
        self.GRETunnels = resource("GRETunnels")
        self.DHCPServers = resource("DHCPServers")
        self.DHCPRelays = resource("DHCPRelays")
        self.DNSProxies = resource("DNSProxies")
        self.GlobalProtectPortals = resource("GlobalProtectPortals")
        self.GlobalProtectGateways = resource("GlobalProtectGateways")
        self.GlobalProtectGatewayAgentTunnels = resource(
            "GlobalProtectGatewayAgentTunnels",
        )
        self.GlobalProtectGatewaySatelliteTunnels = resource(
            "GlobalProtectGatewaySatelliteTunnels",
        )
        self.GlobalProtectGatewayMDMServers = resource("GlobalProtectGatewayMDMServers")
        self.GlobalProtectClientlessApps = resource("GlobalProtectClientlessApps")
        self.GlobalProtectClientlessAppGroups = resource(
            "GlobalProtectClientlessAppGroups",
        )
        self.QoSInterfaces = resource("QoSInterfaces")
        self.LLDP = resource("LLDP")
        self.GlobalProtectIPSecCryptoNetworkProfiles = resource(
            "GlobalProtectIPSecCryptoNetworkProfiles",
        )
        self.IKEGatewayNetworkProfiles = resource("IKEGatewayNetworkProfiles")
        self.IKECryptoNetworkProfiles = resource("IKECryptoNetworkProfiles")
        self.MonitorNetworkProfiles = resource("MonitorNetworkProfiles")
        self.InterfaceManagementNetworkProfiles = resource(
            "InterfaceManagementNetworkProfiles",
        )
        self.ZoneProtectionNetworkProfiles = resource("ZoneProtectionNetworkProfiles")
        self.QoSNetworkProfiles = resource("QoSNetworkProfiles")
        self.LLDPNetworkProfiles = resource("LLDPNetworkProfiles")
        self.SDWANInterfaceProfiles = resource("SDWANInterfaceProfiles")


class PanoramaDevicesResourceType(PanoramaResourceType):
    DEFAULT_PARAMS: ClassVar = {
        "location": "template",
    }
    resource_type = "Device"

    def __init__(self, client, domain, version="v10.1"):
        resource_type = self.resource_type
        super().__init__(client, resource_type)

        def resource(name):
            return PanoramaResource(
                client,
                domain,
                resource_type,
                name,
                version=version,
                default_params=self.DEFAULT_PARAMS,
            )

        self.VirtualSystems = resource("VirtualSystems")
